#include <util/atomic.h>
#include "buttons.h"
#include "crc.h"
#include "data_link_rx.h"
#include "frame.h"
#include "hal_rx.h"


#define FRAME_SIZE (sizeof(frame_t))


typedef struct receiver_t_
{
    data_link_rx_t data_link;
    uint8_t buffer[FRAME_SIZE];
    uint8_t buffer_pointer;
    uint8_t good_frames_counter;
    bool has_signal;
}
receiver_t;


static void receiver_init(receiver_t *receiver)
{
    data_link_rx_init(&receiver->data_link);
    receiver->buffer_pointer = 0;
    receiver->has_signal = false;
    receiver->good_frames_counter = 0;
}


static bool receiver_get_frame(receiver_t *receiver, frame_t *output)
{
    bool result = false;

    while (hal_rx_has_rf_data())
    {
        uint8_t wire_data = hal_rx_receive_rf_data();
        data_link_rx_receive(&receiver->data_link, wire_data);

        if (data_link_rx_get_start_condition(&receiver->data_link))
        {
            receiver->buffer_pointer = 0;
        }

        while (data_link_rx_has_data(&receiver->data_link))
        {
            uint8_t data = data_link_rx_get_data(&receiver->data_link);

            if (receiver->buffer_pointer < FRAME_SIZE)
            {
                receiver->buffer[receiver->buffer_pointer] = data;
                ++receiver->buffer_pointer;
            }

            if ((receiver->buffer_pointer == FRAME_SIZE) && (frame_verify_crc((frame_t *) receiver->buffer)))
            {
                *output = *((frame_t *) receiver->buffer);                    
                result = true;
                ++receiver->buffer_pointer;

                ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
                {
                    ++receiver->good_frames_counter;
                }
            }
        }
    }

    return result;
}


bool receiver_has_signal(receiver_t *receiver)
{
    bool result = false;

    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        result = receiver->has_signal;
    }

    return result;
}


#include <util/delay.h>

static void receiver_compute_stats(receiver_t *receiver, uint16_t time_ms)
{
    if (hal_rx_get_status() == HAL_RX_STATUS_BUFFER_OVERFLOW)
    {
        while (true)
        {
            hal_rx_set_status_led(true);
            _delay_ms(100);
            hal_rx_set_status_led(false);
            _delay_ms(100);
        }
    }

    if (time_ms == 0 || time_ms == 500)
    {
        receiver->has_signal = receiver->good_frames_counter >= 5;
        hal_rx_set_status_led(receiver->has_signal);
        receiver->good_frames_counter = 0;
    }
}


static receiver_t receiver;
static frame_t frame;


int main(void)
{
    hal_rx_init();
    receiver_init(&receiver);
    hal_rx_set_timer_callback((hal_rx_callback_t) &receiver_compute_stats, (void *) &receiver);
    
    while (1)
    {
        bool has_signal = receiver_has_signal(&receiver);
        bool has_new_frame = receiver_get_frame(&receiver, &frame);

        if (!has_signal)
        {
            hal_rx_drive_motor(0);
            hal_rx_drive_servo1(1500);
            hal_rx_drive_servo2(1500);
            hal_rx_set_pwm_output(0);
            hal_rx_set_gp_output(0, false);
            hal_rx_set_gp_output(1, false);
            hal_rx_set_gp_output(2, false);
            hal_rx_set_gp_output(3, false);
            hal_rx_set_gp_output(4, false);
            hal_rx_set_gp_output(5, false);
            hal_rx_set_gp_output(6, false);
        }
        else if (has_new_frame)
        {
            hal_rx_drive_servo1(1500 - 512 * 2 + frame.analog2 * 2);
            hal_rx_drive_servo2(1500 - 512 * 2 + frame.analog4 * 2);
            hal_rx_drive_motor((int16_t) frame.analog3 / 4 - 128);
            hal_rx_set_pwm_output(frame.analog4 / 4);
            hal_rx_set_gp_output(0, (frame.buttons & BUTTON_TRIANGLE) != 0);
            hal_rx_set_gp_output(1, (frame.buttons & BUTTON_CIRCLE) != 0);
            hal_rx_set_gp_output(2, (frame.buttons & BUTTON_X) != 0);
            hal_rx_set_gp_output(3, (frame.buttons & BUTTON_SQUARE) != 0);
            hal_rx_set_gp_output(4, (frame.toggles & BUTTON_LEFT) != 0);
            hal_rx_set_gp_output(5, (frame.toggles & BUTTON_RIGHT) != 0);
            hal_rx_set_gp_output(6, (frame.toggles & BUTTON_DOWN) != 0);
        }
    }

    return 0;
}